import calendar
import datetime
import random
from typing import Any, Dict, Iterable, Optional, Union

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from atlassian_jwt_auth import algorithms, key
from atlassian_jwt_auth.key import BasePrivateKeyRetriever, KeyIdentifier


class JWTAuthSigner(object):
    def __init__(
        self, issuer: str, private_key_retriever: BasePrivateKeyRetriever, **kwargs: Any
    ) -> None:
        self.issuer = issuer
        self.private_key_retriever = private_key_retriever
        self.lifetime = kwargs.get("lifetime", datetime.timedelta(minutes=1))
        self.algorithm = kwargs.get("algorithm", "RS256")
        self.subject = kwargs.get("subject", None)
        self._private_keys_cache: Dict[str, Any] = dict()

        if self.algorithm not in set(algorithms.get_permitted_algorithm_names()):
            raise ValueError("Algorithm, '%s', is not permitted." % self.algorithm)
        if self.lifetime > datetime.timedelta(hours=1):
            raise ValueError(
                "lifetime, '%s',exceeds the allowed 1 hour max" % (self.lifetime)
            )

    def _obtain_private_key(
        self, key_identifier: KeyIdentifier, private_key_pem: Union[str, bytes]
    ):
        """returns a loaded instance of the given private key either from
        cache or from the given private_key_pem.
        """
        priv_key = self._private_keys_cache.get(key_identifier.key_id, None)
        if priv_key is not None:
            return priv_key
        private_key_bytes: bytes
        if not isinstance(private_key_pem, bytes):
            private_key_bytes = private_key_pem.encode()
        else:
            private_key_bytes = private_key_pem
        priv_key = serialization.load_pem_private_key(
            private_key_bytes, password=None, backend=default_backend()
        )
        if len(self._private_keys_cache) > 10:
            self._private_keys_cache = dict()
        self._private_keys_cache[key_identifier.key_id] = priv_key
        return priv_key

    def _generate_claims(
        self, audience: Union[str, Iterable[str]], **kwargs: Any
    ) -> Dict[Any, Any]:
        """returns a new dictionary of claims."""
        now = self._now()
        claims = {
            "iss": self.issuer,
            "exp": now + self.lifetime,
            "iat": now,
            "aud": audience,
            "jti": "%s:%s"
            % (now.strftime("%s"), random.SystemRandom().getrandbits(32)),
            "nbf": now,
            "sub": self.subject or self.issuer,
        }
        claims.update(kwargs.get("additional_claims", {}))
        return claims

    def _now(self) -> datetime.datetime:
        return datetime.datetime.now(datetime.timezone.utc)

    def generate_jwt(self, audience: Union[str, Iterable[str]], **kwargs: Any) -> bytes:
        """returns a new signed jwt for use."""
        key_identifier, private_key_pem = self.private_key_retriever.load(self.issuer)
        private_key = self._obtain_private_key(key_identifier, private_key_pem)
        token: str = jwt.encode(
            self._generate_claims(audience, **kwargs),
            key=private_key,
            algorithm=self.algorithm,
            headers={
                "kid": key_identifier.key_id
                if isinstance(key_identifier, KeyIdentifier)
                else key_identifier
            },
        )
        return token.encode("utf-8")


class TokenReusingJWTAuthSigner(JWTAuthSigner):
    def __init__(
        self, issuer: str, private_key_retriever: BasePrivateKeyRetriever, **kwargs: Any
    ) -> None:
        super(TokenReusingJWTAuthSigner, self).__init__(
            issuer, private_key_retriever, **kwargs
        )
        self.reuse_threshold = kwargs.get("reuse_jwt_threshold", 0.95)

    def get_cached_token(
        self, audience: Union[str, Iterable[str]], **kwargs: Any
    ) -> Optional[bytes]:
        """returns the cached token. If there is no matching cached token
        then None is returned.
        """
        return getattr(self, "_previous_token", None)

    def set_cached_token(self, value: Any) -> None:
        """sets the cached token."""
        self._previous_token = value

    def can_reuse_token(self, existing_token, claims) -> bool:
        """returns True if the provided existing token can be reused
        for the claims provided.
        """
        if existing_token is None:
            return False
        existing_claims = jwt.decode(
            existing_token, options={"verify_signature": False}
        )
        existing_lifetime = int(existing_claims["exp"]) - int(existing_claims["iat"])
        this_lifetime = (claims["exp"] - claims["iat"]).total_seconds()
        if existing_lifetime != this_lifetime:
            return False
        about_to_expire = int(existing_claims["iat"]) + (
            self.reuse_threshold * existing_lifetime
        )
        if calendar.timegm(self._now().utctimetuple()) > about_to_expire:
            return False
        if set(claims.keys()) != set(existing_claims.keys()):
            return False
        for dict_key, val in claims.items():
            if dict_key in ["exp", "iat", "jti", "nbf"]:
                continue
            if existing_claims[dict_key] != val:
                return False
        return True

    def generate_jwt(self, audience: Union[str, Iterable[str]], **kwargs: Any) -> bytes:
        existing_token = self.get_cached_token(audience, **kwargs)
        claims = self._generate_claims(audience, **kwargs)
        if existing_token and self.can_reuse_token(existing_token, claims):
            return existing_token
        token = super(TokenReusingJWTAuthSigner, self).generate_jwt(audience, **kwargs)
        self.set_cached_token(token)
        return token


def _create_signer(
    issuer: str, private_key_retriever: BasePrivateKeyRetriever, **kwargs: Any
) -> JWTAuthSigner:
    signer_cls = JWTAuthSigner
    if kwargs.get("reuse_jwts", None):
        signer_cls = TokenReusingJWTAuthSigner
    return signer_cls(issuer, private_key_retriever, **kwargs)


def create_signer(
    issuer: str,
    key_identifier: Union[KeyIdentifier, str],
    private_key_pem: Union[str, bytes],
    **kwargs: Any,
) -> JWTAuthSigner:
    private_key_retriever = key.StaticPrivateKeyRetriever(
        key_identifier, private_key_pem
    )
    return _create_signer(issuer, private_key_retriever, **kwargs)


def create_signer_from_file_private_key_repository(
    issuer: str, private_key_repository: str, **kwargs: Any
) -> JWTAuthSigner:
    private_key_retriever = key.FilePrivateKeyRetriever(private_key_repository)
    return _create_signer(issuer, private_key_retriever, **kwargs)



from abc import ABCMeta, abstractmethod, abstractproperty
from functools import lru_cache
from typing import Any, Dict, Optional, Union

from atlassian_jwt_auth import HTTPSPublicKeyRetriever, JWTAuthVerifier

from .utils import SettingsDict


@lru_cache(maxsize=20)
def _get_verifier(settings) -> JWTAuthVerifier:
    """This has been extracted out of Backend to avoid possible memory
    leaks via retained instance references.
    """
    retriever = settings.ASAP_KEY_RETRIEVER_CLASS(
        base_url=settings.ASAP_PUBLICKEY_REPOSITORY
    )
    kwargs = {}
    if settings.ASAP_SUBJECT_SHOULD_MATCH_ISSUER is not None:
        kwargs = {
            "subject_should_match_issuer": settings.ASAP_SUBJECT_SHOULD_MATCH_ISSUER
        }
    if settings.ASAP_CHECK_JTI_UNIQUENESS is not None:
        kwargs["check_jti_uniqueness"] = settings.ASAP_CHECK_JTI_UNIQUENESS
    return JWTAuthVerifier(retriever, **kwargs)


class Backend:
    """Abstract class representing a web framework backend

    Backends allow specific implementation details of web frameworks to be
    abstracted away from the underlying logic of ASAP.
    """

    __metaclass__ = ABCMeta

    default_headers_401 = {"WWW-Authenticate": "Bearer"}
    default_settings = {
        # The class to be instantiated to retrieve public keys
        "ASAP_KEY_RETRIEVER_CLASS": HTTPSPublicKeyRetriever,
        # The repository URL where the key retriever can fetch public keys
        "ASAP_PUBLICKEY_REPOSITORY": None,
        # Whether or not ASAP authentication is required
        # This is primarily useful when phasing in ASAP authentication
        "ASAP_REQUIRED": True,
        # The valid audience value expected when authenticating tokens
        "ASAP_VALID_AUDIENCE": None,
        # The amount of leeway to apply when evaluating token expiration
        # timestamps
        "ASAP_VALID_LEEWAY": 0,
        # An iterable of valid token issuers allowed to authenticate
        # (this can be overridden at the decorator level)
        "ASAP_VALID_ISSUERS": None,
        # Enforce that the ASAP subject must match the issuer
        "ASAP_SUBJECT_SHOULD_MATCH_ISSUER": None,
        # Enforce that tokens have a unique JTI
        # Set this to True to enforce JTI uniqueness checking.
        "ASAP_CHECK_JTI_UNIQUENESS": None,
    }

    @abstractmethod
    def get_authorization_header(self, request: Optional[Any] = None) -> bytes:
        pass

    @abstractmethod
    def get_401_response(
        self,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        request: Optional[Any] = None,
    ) -> Any:
        pass

    @abstractmethod
    def get_403_response(
        self,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        request: Optional[Any] = None,
    ) -> Any:
        pass

    @abstractmethod
    def set_asap_claims_for_request(self, request: Any, claims: Any) -> None:
        pass

    @abstractproperty
    def settings(self) -> SettingsDict:
        return SettingsDict(self.default_settings)

    def get_asap_token(self, request: Any) -> Optional[bytes]:
        auth_header = self.get_authorization_header(request)

        if auth_header is None:
            return None

        if isinstance(auth_header, str):
            # Per PEP-3333, headers must be in ISO-8859-1 or use an RFC-2047
            # MIME encoding. We don't really care about MIME encoded
            # headers, but some libraries allow sending bytes (Django tests)
            # and some (requests) always send str so we need to convert if
            # that is the case to properly support Python 3.
            auth_header = auth_header.encode(encoding="iso-8859-1")

        auth_values = auth_header.split(b" ")
        if len(auth_values) != 2 or auth_values[0].lower() != b"bearer":
            return None

        return auth_values[1]

    def get_verifier(self, settings: Optional[SettingsDict] = None) -> JWTAuthVerifier:
        """Returns a verifier for ASAP JWT tokens"""
        if settings is None:
            settings = self.settings
        return self._get_verifier(settings)

    def _get_verifier(self, settings: SettingsDict) -> JWTAuthVerifier:
        return _get_verifier(settings)

    def _process_settings(self, settings: Union[SettingsDict, Dict]) -> SettingsDict:
        valid_issuers = settings.get("ASAP_VALID_ISSUERS")
        if valid_issuers:
            settings["ASAP_VALID_ISSUERS"] = set(valid_issuers)

        valid_aud = settings.get("ASAP_VALID_AUDIENCE")
        if valid_aud and isinstance(valid_aud, list):
            settings["ASAP_VALID_AUDIENCE"] = set(valid_aud)

        return SettingsDict(settings)



from functools import wraps
from typing import Any, Callable, Dict, Iterable, Optional

from jwt.exceptions import InvalidIssuerError, InvalidTokenError

from .asap import _process_asap_token, _verify_issuers
from .backend import Backend
from .utils import SettingsDict


def _with_asap(
    func: Optional[Callable] = None,
    backend: Optional[Backend] = None,
    issuers: Optional[Iterable[str]] = None,
    required: bool = True,
    subject_should_match_issuer: Optional[bool] = None,
) -> Callable:
    if backend is None:
        raise ValueError("Invalid value for backend. Use a subclass instead.")

    def with_asap_decorator(func: Callable):
        @wraps(func)
        def with_asap_wrapper(*args: Any, **kwargs: Any) -> Optional[str]:
            settings = _update_settings_from_kwargs(
                backend.settings,
                issuers=issuers,
                required=required,
                subject_should_match_issuer=subject_should_match_issuer,
            )

            request = None
            if len(args) > 0:
                request = args[0]

            error_response = _process_asap_token(request, backend, settings)

            if error_response is not None:
                return error_response

            return func(*args, **kwargs)

        return with_asap_wrapper

    if callable(func):
        return with_asap_decorator(func)

    return with_asap_decorator


def _restrict_asap(
    func: Optional[Callable] = None,
    backend: Optional[Backend] = None,
    issuers: Optional[Iterable[str]] = None,
    required: bool = True,
    subject_should_match_issuer: Optional[bool] = None,
):
    """Decorator to allow endpoint-specific ASAP authorization, assuming ASAP
    authentication has already occurred.
    """

    def restrict_asap_decorator(func: Callable) -> Optional[Any]:
        @wraps(func)
        def restrict_asap_wrapper(request, *args, **kwargs) -> Any:
            if backend is None:
                raise ValueError("Backend cannot be None")
            settings = _update_settings_from_kwargs(
                backend.settings,
                issuers=issuers,
                required=required,
                subject_should_match_issuer=subject_should_match_issuer,
            )
            asap_claims = getattr(request, "asap_claims", None)
            error_response = None

            if not asap_claims:
                if required:
                    return backend.get_401_response("Unauthorized", request=request)
                else:
                    # Claims are not required and asap claims are not present.
                    return func(request, *args, **kwargs)
            try:
                _verify_issuers(asap_claims, settings.ASAP_VALID_ISSUERS)
            except InvalidIssuerError:
                error_response = backend.get_403_response(
                    "Forbidden: Invalid token issuer", request=request
                )
            except InvalidTokenError:
                error_response = backend.get_401_response(
                    "Unauthorized: Invalid token", request=request
                )

            if error_response and required:
                return error_response

            return func(request, *args, **kwargs)

        return restrict_asap_wrapper

    if callable(func):
        return restrict_asap_decorator(func)

    return restrict_asap_decorator


def _update_settings_from_kwargs(
    settings: Dict[Any, Any],
    issuers: Optional[Iterable] = None,
    required: bool = True,
    subject_should_match_issuer: Optional[bool] = None,
) -> SettingsDict:
    settings = settings.copy()

    if issuers is not None:
        settings["ASAP_VALID_ISSUERS"] = set(issuers)

    if required is not None:
        settings["ASAP_REQUIRED"] = required

    if subject_should_match_issuer is not None:
        settings["ASAP_SUBJECT_SHOULD_MATCH_ISSUER"] = subject_should_match_issuer

    return SettingsDict(settings)



import warnings

from .decorators import requires_asap

__all__ = [
    "requires_asap",
]

warnings.warn(
    "The atlassian_jwt_auth.contrib.flask_app package is deprecated in 4.0.0 "
    "in favour of atlassian_jwt_auth.frameworks.flask.",
    DeprecationWarning,
    stacklevel=2,
)



import asyncio
from typing import Any, Dict, Iterable, Union

import jwt

from atlassian_jwt_auth import key
from atlassian_jwt_auth.verifier import JWTAuthVerifier as _JWTAuthVerifier


class JWTAuthVerifier(_JWTAuthVerifier):  # type: ignore[override]
    async def verify_jwt(  # type: ignore[override]
        self,
        a_jwt: str,
        audience: Union[str, Iterable[str]],
        leeway: int = 0,
        **requests_kwargs: Any,
    ) -> Dict[Any, Any]:
        """Verify if the token is correct

        Returns:
             dict: the claims of the given jwt if verification is successful.

        Raises:
            ValueError: if verification failed.
        """
        key_identifier = key._get_key_id_from_jwt_header(a_jwt)

        public_key = self._retrieve_pub_key(key_identifier, requests_kwargs)
        if asyncio.iscoroutine(public_key):
            public_key = await public_key

        alg = jwt.get_unverified_header(a_jwt).get("alg", None)
        public_key_obj = self._load_public_key(public_key, alg)
        return self._decode_jwt(
            a_jwt, key_identifier, public_key_obj, audience=audience, leeway=leeway
        )



from typing import Any, Optional

from ..common.backend import Backend
from ..common.utils import SettingsDict


class WSGIBackend(Backend):
    def __init__(self, settings) -> None:
        self._settings = SettingsDict(settings)

    def get_authorization_header(self, request=None):
        if request is None:
            raise ValueError("No request available")

        return request.environ.get("HTTP_AUTHORIZATION", b"")

    def get_401_response(
        self,
        data: Any = None,
        headers: Optional[Any] = None,
        request: Optional[Any] = None,
    ) -> str:
        if request is None:
            raise TypeError("request must have a value")

        if headers is None:
            headers = {}

        headers.update(self.default_headers_401)

        request.start_response("401 Unauthorized", list(headers.items()), None)
        return ""

    def get_403_response(
        self,
        data: Any = None,
        headers: Optional[Any] = None,
        request: Optional[Any] = None,
    ) -> str:
        if request is None:
            raise TypeError("request must have a value")

        if headers is None:
            headers = {}

        request.start_response("403 Forbidden", list(headers.items()), None)
        return ""

    def set_asap_claims_for_request(self, request: Any, claims: Any) -> None:
        request.environ["ATL_ASAP_CLAIMS"] = claims

    @property
    def settings(self) -> SettingsDict:
        settings = {}
        settings.update(self.default_settings)

        for k in settings.keys():
            value = getattr(self._settings, k, None)
            if value is None:
                continue

            settings[k] = value

        return self._process_settings(settings)



from django.urls import path

from atlassian_jwt_auth.frameworks.django.tests import views

urlpatterns = [
    path("asap/expected", views.expected_view, name="expected"),
    path(r"^asap/unexpected", views.unexpected_view, name="unexpected"),
    path("^asap/decorated", views.decorated_view, name="decorated"),
    path("asap/settings", views.settings_view, name="settings"),
    path(
        "asap/subject_does_not_need_to_match_issuer",
        views.subject_does_not_need_to_match_issuer_view,
        name="subject_does_not_need_to_match_issuer",
    ),
    path(
        "asap/subject_does_need_to_match_issuer_view",
        views.subject_does_need_to_match_issuer_view,
        name="subject_does_need_to_match_issuer",
    ),
    path(
        "asap/subject_does_not_need_to_match_issuer_from_settings",
        views.subject_does_not_need_to_match_issuer_from_settings_view,
        name="subject_does_not_need_to_match_issuer_from_settings",
    ),
    path("asap/needed", views.needed_view, name="needed"),
    path(r"asap/unneeded", views.unneeded_view, name="unneeded"),
    path(
        r"asap/restricted_issuer",
        views.restricted_issuer_view,
        name="restricted_issuer",
    ),
    path(
        "asap/restricted_subject",
        views.restricted_subject_view,
        name="restricted_subject",
    ),
]



from django.http import HttpResponse

from atlassian_jwt_auth.contrib.django.decorators import requires_asap, validate_asap
from atlassian_jwt_auth.frameworks.django import restrict_asap, with_asap


@with_asap(issuers=["client-app"])
def expected_view(request):
    return HttpResponse("Greatest Success!")


@with_asap(issuers=["unexpected"])
def unexpected_view(request):
    return HttpResponse("This should fail.")


@with_asap(issuers=["whitelist"])
def decorated_view(request):
    return HttpResponse("Only the right issuer is allowed.")


@requires_asap()
def settings_view(request):
    return HttpResponse("Any settings issuer is allowed.")


@with_asap(subject_should_match_issuer=False)
def subject_does_not_need_to_match_issuer_view(request):
    return HttpResponse("Subject does not need to match issuer.")


@with_asap(subject_should_match_issuer=True)
def subject_does_need_to_match_issuer_view(request):
    return HttpResponse("Subject does need to match issuer.")


@with_asap()
def subject_does_not_need_to_match_issuer_from_settings_view(request):
    return HttpResponse("Subject does not need to match issuer (settings).")


@restrict_asap
def needed_view(request):
    return HttpResponse("one")


@restrict_asap(required=False)
def unneeded_view(request):
    return HttpResponse("two")


@restrict_asap(issuers=["whitelist"])
def restricted_issuer_view(request):
    return HttpResponse("three")


@validate_asap(subjects=["client-app"])
def restricted_subject_view(request):
    return HttpResponse("four")



from atlassian_jwt_auth.algorithms import get_permitted_algorithm_names
from atlassian_jwt_auth.key import (
    HTTPSPublicKeyRetriever,
    KeyIdentifier,
)
from atlassian_jwt_auth.signer import (
    create_signer,
    create_signer_from_file_private_key_repository,
)
from atlassian_jwt_auth.verifier import JWTAuthVerifier

__all__ = [
    "get_permitted_algorithm_names",
    "HTTPSPublicKeyRetriever",
    "KeyIdentifier",
    "create_signer",
    "create_signer_from_file_private_key_repository",
    "JWTAuthVerifier",
]



import os

import django
from django.test.testcases import SimpleTestCase
from django.test.utils import modify_settings, override_settings

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from atlassian_jwt_auth.contrib.tests.utils import (
    get_static_retriever_class,
)
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.tests.utils import (
    RS256KeyTestMixin,
    create_token,
)


class DjangoAsapMixin(object):
    @classmethod
    def setUpClass(cls):
        os.environ.setdefault(
            "DJANGO_SETTINGS_MODULE",
            "atlassian_jwt_auth.frameworks.django.tests.settings",
        )

        django.setup()
        super(DjangoAsapMixin, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        super(DjangoAsapMixin, cls).tearDownClass()
        del os.environ["DJANGO_SETTINGS_MODULE"]

    def setUp(self):
        super(DjangoAsapMixin, self).setUp()
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )

        self.retriever = get_static_retriever_class(
            {"client-app/key01": self._public_key_pem}
        )

        self.test_settings = {"ASAP_KEY_RETRIEVER_CLASS": self.retriever}


@modify_settings(
    MIDDLEWARE={
        "prepend": "atlassian_jwt_auth.frameworks.django.asap_middleware",
    }
)
class TestAsapMiddleware(DjangoAsapMixin, RS256KeyTestMixin, SimpleTestCase):
    def check_response(
        self,
        view_name,
        response_content="",
        status_code=200,
        issuer="client-app",
        audience="server-app",
        key_id="client-app/key01",
        subject=None,
        private_key=None,
        token=None,
        authorization=None,
        retriever_key=None,
    ) -> None:
        if authorization is None:
            if token is None:
                if private_key is None:
                    private_key = self._private_key_pem
                token = create_token(
                    issuer=issuer,
                    audience=audience,
                    key_id=key_id,
                    private_key=private_key,
                    subject=subject,
                )
            authorization = b"Bearer " + token

        test_settings = self.test_settings.copy()
        if retriever_key is not None:
            retriever = get_static_retriever_class(
                {retriever_key: self._public_key_pem}
            )
            test_settings["ASAP_KEY_RETRIEVER_CLASS"] = retriever

        with override_settings(**test_settings):
            response = self.client.get(
                reverse(view_name), HTTP_AUTHORIZATION=authorization
            )

        self.assertContains(response, response_content, status_code=status_code)

    def test_request_with_valid_token_is_allowed(self):
        self.check_response("needed", "one", 200)

    def test_request_with_valid_token_multiple_allowed_auds(self):
        audiences = ["server-app", "another_one"]
        self.test_settings["ASAP_VALID_AUDIENCE"] = audiences
        for aud in audiences:
            self.check_response("needed", "one", 200, audience=aud)

    def test_request_with_valid_token_multiple_allowed_auds_invalid_aud(self):
        audiences = ["server-app", "another_one"]
        self.test_settings["ASAP_VALID_AUDIENCE"] = audiences
        self.check_response("needed", "Unauthorized", 401, audience="invalid")

    def test_request_with_duplicate_jti_is_rejected_as_per_setting(self):
        self.test_settings["ASAP_CHECK_JTI_UNIQUENESS"] = True
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        str_auth = "Bearer " + token.decode(encoding="iso-8859-1")
        self.check_response("needed", "one", 200, authorization=str_auth)
        self.check_response("needed", "duplicate jti", 401, authorization=str_auth)

    def _assert_request_with_duplicate_jti_is_accepted(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        str_auth = "Bearer " + token.decode(encoding="iso-8859-1")
        self.check_response("needed", "one", 200, authorization=str_auth)
        self.check_response("needed", "one", 200, authorization=str_auth)

    def test_request_with_duplicate_jti_is_accepted(self):
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_duplicate_jti_is_accepted_as_per_setting(self):
        self.test_settings["ASAP_CHECK_JTI_UNIQUENESS"] = False
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_string_headers_is_allowed(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        str_auth = "Bearer " + token.decode(encoding="iso-8859-1")
        self.check_response("needed", "one", 200, authorization=str_auth)

    def test_request_with_invalid_audience_is_rejected(self):
        self.check_response("needed", "Unauthorized", 401, audience="invalid")

    def test_request_with_invalid_token_is_rejected(self):
        self.check_response(
            "needed", "Unauthorized", 401, authorization="Bearer invalid"
        )

    def test_request_without_token_is_rejected(self):
        with override_settings(**self.test_settings):
            response = self.client.get(reverse("needed"))

        self.assertContains(response, "Unauthorized", status_code=401)

    def test_request_with_invalid_issuer_is_rejected(self):
        self.check_response(
            "needed",
            "Forbidden",
            403,
            issuer="something-invalid",
            key_id="something-invalid/key01",
            retriever_key="something-invalid/key01",
        )

    def test_request_non_whitelisted_decorated_issuer_is_rejected(self):
        self.check_response(
            "needed",
            "Forbidden",
            403,
            issuer="unexpected",
            key_id="unexpected/key01",
            retriever_key="unexpected/key01",
        )

    def test_request_non_decorated_issuer_is_rejected(self):
        self.check_response("restricted_issuer", "Forbidden", 403)

    def test_request_decorated_issuer_is_allowed(self):
        self.check_response(
            "restricted_issuer",
            "three",
            issuer="whitelist",
            key_id="whitelist/key01",
            retriever_key="whitelist/key01",
        )

    # TODO: modify JWTAuthSigner to allow non-issuer subjects and update the
    # decorated subject test cases
    def test_request_non_decorated_subject_is_rejected(self):
        self.check_response(
            "restricted_subject",
            "Forbidden",
            403,
            issuer="whitelist",
            key_id="whitelist/key01",
            retriever_key="whitelist/key01",
        )

    def test_request_using_settings_only_is_allowed(self):
        self.check_response("unneeded", "two")

    def test_request_subject_does_not_need_to_match_issuer_from_settings(self) -> None:
        self.test_settings["ASAP_SUBJECT_SHOULD_MATCH_ISSUER"] = False
        self.check_response("needed", "one", 200, subject="different_than_is")

    def test_request_subject_and_issue_not_matching(self):
        self.check_response(
            "needed",
            "Subject and Issuer do not match",
            401,
            subject="different_than_is",
        )


class TestAsapDecorator(DjangoAsapMixin, RS256KeyTestMixin, SimpleTestCase):
    def test_request_with_valid_token_is_allowed(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse("expected"), HTTP_AUTHORIZATION=b"Bearer " + token
            )

        self.assertContains(response, "Greatest Success!", status_code=200)

    def test_request_with_string_headers_is_allowed(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        str_token = token.decode(encoding="iso-8859-1")
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse("expected"), HTTP_AUTHORIZATION="Bearer " + str_token
            )

        self.assertContains(response, "Greatest Success!", status_code=200)

    def test_request_with_invalid_audience_is_rejected(self):
        token = create_token(
            issuer="client-app",
            audience="something-invalid",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse("expected"), HTTP_AUTHORIZATION=b"Bearer " + token
            )

        self.assertContains(response, "Unauthorized: Invalid token", status_code=401)

    def test_request_with_invalid_token_is_rejected(self):
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse("expected"), HTTP_AUTHORIZATION=b"Bearer notavalidtoken"
            )

        self.assertContains(response, "Unauthorized: Invalid token", status_code=401)

    def test_request_without_token_is_rejected(self):
        with override_settings(**self.test_settings):
            response = self.client.get(reverse("expected"))

        self.assertContains(response, "Unauthorized", status_code=401)

    def test_request_with_invalid_issuer_is_rejected(self):
        retriever = get_static_retriever_class(
            {"something-invalid/key01": self._public_key_pem}
        )
        token = create_token(
            issuer="something-invalid",
            audience="server-app",
            key_id="something-invalid/key01",
            private_key=self._private_key_pem,
        )
        with override_settings(ASAP_KEY_RETRIEVER_CLASS=retriever):
            response = self.client.get(
                reverse("expected"), HTTP_AUTHORIZATION=b"Bearer " + token
            )

        self.assertContains(
            response, "Forbidden: Invalid token issuer", status_code=403
        )

    def test_request_non_decorated_issuer_is_rejected(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse("decorated"), HTTP_AUTHORIZATION=b"Bearer " + token
            )

        self.assertContains(
            response, "Forbidden: Invalid token issuer", status_code=403
        )

    def test_request_decorated_issuer_is_allowed(self):
        retriever = get_static_retriever_class(
            {"whitelist/key01": self._public_key_pem}
        )
        token = create_token(
            issuer="whitelist",
            audience="server-app",
            key_id="whitelist/key01",
            private_key=self._private_key_pem,
        )
        with override_settings(ASAP_KEY_RETRIEVER_CLASS=retriever):
            response = self.client.get(
                reverse("decorated"), HTTP_AUTHORIZATION=b"Bearer " + token
            )

        self.assertContains(response, "Only the right issuer is allowed.")

    def test_request_using_settings_only_is_allowed(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
        )
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse("settings"), HTTP_AUTHORIZATION=b"Bearer " + token
            )

        self.assertContains(response, "Any settings issuer is allowed.")

    def test_request_subject_does_not_need_to_match_issuer(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
            subject="not-client-app",
        )
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse("subject_does_not_need_to_match_issuer"),
                HTTP_AUTHORIZATION=b"Bearer " + token,
            )

        self.assertContains(response, "Subject does not need to match issuer.")

    def test_request_subject_does_need_to_match_issuer_override_settings(self):
        """tests that the with_asap decorator can override the
        ASAP_SUBJECT_SHOULD_MATCH_ISSUER setting.
        """
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
            subject="not-client-app",
        )
        with override_settings(
            **dict(self.test_settings, ASAP_SUBJECT_SHOULD_MATCH_ISSUER=False)
        ):
            response = self.client.get(
                reverse("subject_does_need_to_match_issuer"),
                HTTP_AUTHORIZATION=b"Bearer " + token,
            )
            self.assertContains(
                response,
                "Unauthorized: Subject and Issuer do not match",
                status_code=401,
            )

    def test_request_subject_does_not_need_to_match_issuer_from_settings(self):
        token = create_token(
            issuer="client-app",
            audience="server-app",
            key_id="client-app/key01",
            private_key=self._private_key_pem,
            subject="not-client-app",
        )
        with override_settings(
            **dict(self.test_settings, ASAP_SUBJECT_SHOULD_MATCH_ISSUER=False)
        ):
            response = self.client.get(
                reverse("subject_does_not_need_to_match_issuer_from_settings"),
                HTTP_AUTHORIZATION=b"Bearer " + token,
            )

        self.assertContains(
            response, "Subject does not need to match issuer (settings)."
        )



from .decorators import restrict_asap, with_asap
from .middleware import OldStyleASAPMiddleware, asap_middleware

__all__ = ["restrict_asap", "with_asap", "OldStyleASAPMiddleware", "asap_middleware"]



import unittest

import atlassian_jwt_auth


class TestKeyModule(unittest.TestCase):
    """tests for the key module."""

    def test_key_identifier_with_invalid_keys(self) -> None:
        """test that invalid key identifiers are not permitted."""
        keys = [
            "../aha",
            "/a",
            r"\c:a",
            "lk2j34/#$",
            "a../../a",
            "a/;a",
            " ",
            " / ",
            " /",
            "dir/some\0thing",
            "a/#a",
            "a/a?x",
            "a/a;",
        ]
        for key in keys:
            with self.assertRaises(ValueError):
                atlassian_jwt_auth.KeyIdentifier(identifier=key)

    def test_key_identifier_with_valid_keys(self) -> None:
        """test that valid keys work as expected."""
        for key in ["oa.oo/a", "oo.sasdf.asdf/yes", "oo/o"]:
            key_id = atlassian_jwt_auth.KeyIdentifier(identifier=key)
            self.assertEqual(key_id.key_id, key)



"""Provide asyncio support"""

from .auth import JWTAuth
from .key import HTTPSPublicKeyRetriever
from .verifier import JWTAuthVerifier

__all__ = [
    "JWTAuth",
    "HTTPSPublicKeyRetriever",
    "JWTAuthVerifier",
]



import unittest

from flask import Flask

from atlassian_jwt_auth.contrib.flask_app import requires_asap
from atlassian_jwt_auth.contrib.tests.utils import get_static_retriever_class
from atlassian_jwt_auth.frameworks.flask import with_asap
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.tests.utils import (
    create_token,
)


def get_app():
    app = Flask(__name__)
    app.config.update(
        {
            "ASAP_VALID_AUDIENCE": "server-app",
            "ASAP_VALID_ISSUERS": ("client-app",),
            "ASAP_PUBLICKEY_REPOSITORY": None,
        }
    )

    @app.route("/")
    @requires_asap
    def view():
        return "OK"

    @app.route("/restricted-to-another-client/")
    @with_asap(issuers=["another-client"])
    def view_for_another_client_app():
        return "OK"

    return app


class FlaskTests(utils.RS256KeyTestMixin, unittest.TestCase):
    """tests for the atlassian_jwt_auth.contrib.tests.flask"""

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )

        self.app = get_app()
        self.client = self.app.test_client()

        retriever = get_static_retriever_class(
            {"client-app/key01": self._public_key_pem}
        )
        self.app.config["ASAP_KEY_RETRIEVER_CLASS"] = retriever

    def send_request(self, token, url="/"):
        """returns the response of sending a request containing the given
        token sent in the Authorization header.
        """

        # Note: We send the auth header as a string and not bytes here
        # due to how Werkzeug's Header code works.
        return self.client.get(
            url, headers={"Authorization": (b"Bearer " + token).decode("iso-8859-1")}
        )

    def test_request_with_valid_token_is_allowed(self):
        token = create_token(
            "client-app", "server-app", "client-app/key01", self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 200)

    def test_request_with_valid_token_multiple_allowed_auds(self):
        audiences = ["server-app", "another_one"]
        self.app.config["ASAP_VALID_AUDIENCE"] = audiences
        for aud in audiences:
            token = create_token(
                "client-app", aud, "client-app/key01", self._private_key_pem
            )
            self.assertEqual(self.send_request(token).status_code, 200)

    def test_request_with_valid_token_multiple_allowed_auds_invalid_aud(self):
        audiences = ["server-app", "another_one"]
        self.app.config["ASAP_VALID_AUDIENCE"] = audiences
        token = create_token(
            "client-app", "invalid", "client-app/key01", self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 401)

    def test_request_with_duplicate_jti_is_rejected_as_per_setting(self):
        self.app.config["ASAP_CHECK_JTI_UNIQUENESS"] = True
        token = create_token(
            "client-app", "server-app", "client-app/key01", self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 200)
        self.assertEqual(self.send_request(token).status_code, 401)

    def _assert_request_with_duplicate_jti_is_accepted(self):
        token = create_token(
            "client-app", "server-app", "client-app/key01", self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 200)
        self.assertEqual(self.send_request(token).status_code, 200)

    def test_request_with_duplicate_jti_is_accepted(self):
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_duplicate_jti_is_accepted_as_per_setting(self):
        self.app.config["ASAP_CHECK_JTI_UNIQUENESS"] = False
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_invalid_audience_is_rejected(self):
        token = create_token(
            "client-app", "invalid-audience", "client-app/key01", self._private_key_pem
        )
        self.assertEqual(self.send_request(token).status_code, 401)

    def test_request_with_invalid_token_is_rejected(self):
        response = self.send_request(b"notavalidtoken")
        self.assertEqual(response.status_code, 401)

    def test_request_with_invalid_issuer_is_rejected(self):
        # Try with a different audience with a valid signature
        self.app.config["ASAP_KEY_RETRIEVER_CLASS"] = get_static_retriever_class(
            {"another-client/key01": self._public_key_pem}
        )
        token = create_token(
            "another-client",
            "server-app",
            "another-client/key01",
            self._private_key_pem,
        )
        self.assertEqual(self.send_request(token).status_code, 403)

    def test_decorated_request_with_invalid_issuer_is_rejected(self):
        # Try with a different audience with a valid signature
        token = create_token(
            "client-app", "server-app", "client-app/key01", self._private_key_pem
        )
        url = "/restricted-to-another-client/"
        self.assertEqual(self.send_request(token, url=url).status_code, 403)

    def test_request_subject_and_issue_not_matching(self):
        token = create_token(
            "client-app",
            "server-app",
            "client-app/key01",
            self._private_key_pem,
            subject="different",
        )
        self.assertEqual(self.send_request(token).status_code, 401)

    def test_request_subject_does_not_need_to_match_issuer_from_settings(self):
        self.app.config["ASAP_SUBJECT_SHOULD_MATCH_ISSUER"] = False
        token = create_token(
            "client-app",
            "server-app",
            "client-app/key01",
            self._private_key_pem,
            subject="different",
        )
        self.assertEqual(self.send_request(token).status_code, 200)



import unittest
from typing import Any, Type

from atlassian_jwt_auth.auth import BaseJWTAuth
from atlassian_jwt_auth.contrib.aiohttp.auth import JWTAuth, create_jwt_auth
from atlassian_jwt_auth.contrib.tests import test_requests
from atlassian_jwt_auth.tests import utils


class BaseAuthTest(test_requests.BaseRequestsTest):
    """tests for the contrib.aiohttp.JWTAuth class"""

    auth_cls: Type[JWTAuth] = JWTAuth

    def _get_auth_header(self, auth) -> bytes:
        return auth.encode().encode("latin1")

    def create_jwt_auth(self, *args: Any, **kwargs: Any) -> BaseJWTAuth:
        return create_jwt_auth(*args, **kwargs)


class RequestsRS256Test(BaseAuthTest, utils.RS256KeyTestMixin, unittest.TestCase):
    pass


class RequestsES256Test(BaseAuthTest, utils.ES256KeyTestMixin, unittest.TestCase):
    pass



import asyncio
import urllib.parse
from asyncio import AbstractEventLoop
from typing import Any, Awaitable, Dict, Optional

import aiohttp

from atlassian_jwt_auth.exceptions import PublicKeyRetrieverException
from atlassian_jwt_auth.key import PEM_FILE_TYPE
from atlassian_jwt_auth.key import HTTPSPublicKeyRetriever as _HTTPSPublicKeyRetriever


class HTTPSPublicKeyRetriever(_HTTPSPublicKeyRetriever):
    """A class for retrieving JWT public keys with aiohttp"""

    _class_session = None

    def __init__(
        self, base_url: str, *, loop: Optional[AbstractEventLoop] = None
    ) -> None:
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        super().__init__(base_url)

    def _get_session(self) -> aiohttp.ClientSession:  # type: ignore[override]
        if HTTPSPublicKeyRetriever._class_session is None:
            HTTPSPublicKeyRetriever._class_session = aiohttp.ClientSession(
                loop=self.loop
            )
        return HTTPSPublicKeyRetriever._class_session

    def _convert_proxies_to_proxy_arg(
        self, url: str, requests_kwargs: Dict[Any, Any]
    ) -> Dict[str, Any]:
        """returns a modified requests_kwargs dict that contains proxy
        information in a form that aiohttp accepts
        (it wants proxy information instead of a dict of proxies).
        """
        proxy = None
        if "proxies" in requests_kwargs:
            scheme = urllib.parse.urlparse(url).scheme
            proxy = requests_kwargs["proxies"].get(scheme, None)
            del requests_kwargs["proxies"]
            requests_kwargs["proxy"] = proxy
        return requests_kwargs

    async def _retrieve(
        self, url: str, requests_kwargs: Dict[Any, Any]
    ) -> Awaitable[str]:
        requests_kwargs = self._convert_proxies_to_proxy_arg(url, requests_kwargs)
        try:
            resp = await self._session.get(
                url,
                headers={
                    "accept":  # type: ignore[misc]
                    PEM_FILE_TYPE
                },
                **requests_kwargs,
            )
            resp.raise_for_status()
            self._check_content_type(url, resp.headers["content-type"])
            return await resp.text()
        except aiohttp.ClientError as e:
            status_code = getattr(e, "code", None)
            raise PublicKeyRetrieverException(e, status_code=status_code)



from collections.abc import Callable
from functools import wraps
from typing import Iterable, Optional

from django.http.response import HttpResponse

from atlassian_jwt_auth.frameworks.django.decorators import with_asap


def validate_asap(
    issuers: Optional[Iterable[str]] = None,
    subjects: Optional[Iterable[str]] = None,
    required: bool = True,
) -> Callable:
    """Decorator to allow endpoint-specific ASAP authorization, assuming ASAP
    authentication has already occurred.

    :param list issuers: A list of issuers that are allowed to use the
        endpoint.
    :param list subjects: A list of subjects that are allowed to use the
        endpoint.
    :param boolean required: Whether or not to require ASAP on this endpoint.
        Note that requirements will be still be verified if claims are present.
    """

    def validate_asap_decorator(func):
        @wraps(func)
        def validate_asap_wrapper(request, *args, **kwargs):
            asap_claims = getattr(request, "asap_claims", None)
            if required and not asap_claims:
                message = "Unauthorized: Invalid or missing token"
                response = HttpResponse(message, status=401)
                response["WWW-Authenticate"] = "Bearer"
                return response

            if asap_claims:
                iss = asap_claims["iss"]
                if issuers and iss not in issuers:
                    message = "Forbidden: Invalid token issuer"
                    return HttpResponse(message, status=403)

                sub = asap_claims.get("sub")
                if subjects and sub not in subjects:
                    message = "Forbidden: Invalid token subject"
                    return HttpResponse(message, status=403)

            return func(request, *args, **kwargs)

        return validate_asap_wrapper

    return validate_asap_decorator


def requires_asap(
    issuers: Optional[Iterable[str]] = None,
    subject_should_match_issuer: Optional[bool] = None,
    func: Optional[Callable] = None,
) -> Callable:
    """Decorator for Django endpoints to require ASAP

    :param list issuers: *required The 'iss' claims that this endpoint is from.
    """
    return with_asap(
        func=func,
        required=True,
        issuers=issuers,
        subject_should_match_issuer=subject_should_match_issuer,
    )



import base64
import logging
import os
import re
from email.message import EmailMessage
from typing import Any, Generator, Iterable, Tuple, Union
from urllib.parse import unquote_plus

import cachecontrol
import cryptography.hazmat.backends
import jwt
import requests
import requests.utils
from cryptography.hazmat.primitives import serialization
from requests.exceptions import ConnectionError, RequestException

from atlassian_jwt_auth.exceptions import (
    KeyIdentifierException,
    PrivateKeyRetrieverException,
    PublicKeyRetrieverException,
)

PEM_FILE_TYPE = "application/x-pem-file"


class KeyIdentifier(object):
    """This class represents a key identifier"""

    def __init__(self, identifier: str) -> None:
        self.__key_id = validate_key_identifier(identifier)

    @property
    def key_id(self) -> str:
        return self.__key_id


def validate_key_identifier(identifier: str) -> str:
    """returns a validated key identifier."""
    regex = re.compile(r"^[\w.\-\+/]*$")
    _error_msg = "Invalid key identifier %s" % identifier
    if not identifier:
        raise KeyIdentifierException(_error_msg)
    if not regex.match(identifier):
        raise KeyIdentifierException(_error_msg)
    normalised = os.path.normpath(identifier)
    if normalised != identifier:
        raise KeyIdentifierException(_error_msg)
    if normalised.startswith("/"):
        raise KeyIdentifierException(_error_msg)
    if ".." in normalised:
        raise KeyIdentifierException(_error_msg)
    return identifier


def _get_key_id_from_jwt_header(a_jwt: Union[str, bytes]) -> KeyIdentifier:
    """returns the key identifier from a jwt header."""
    header = jwt.get_unverified_header(a_jwt)
    return KeyIdentifier(header["kid"])


class BasePublicKeyRetriever(object):
    """Base class for retrieving a public key."""

    def retrieve(self, key_identifier: Union[KeyIdentifier, str], **kwargs) -> Any:
        raise NotImplementedError()


class HTTPSPublicKeyRetriever(BasePublicKeyRetriever):
    """This class retrieves public key from a https location based upon the
    given key id.
    """

    # Use a static requests session, reused/shared by all instances of
    # HTTPSPublicKeyRetriever:
    _class_session = None

    def __init__(self, base_url: str) -> None:
        if base_url is None or not base_url.startswith("https://"):
            raise PublicKeyRetrieverException("The base url must start with https://")
        if not base_url.endswith("/"):
            base_url += "/"
        self.base_url = base_url
        self._session = self._get_session()
        self._proxies = requests.utils.get_environ_proxies(self.base_url)

    def _get_session(self) -> requests.Session:
        if HTTPSPublicKeyRetriever._class_session is None:
            session = cachecontrol.CacheControl(requests.Session())
            session.trust_env = False
            HTTPSPublicKeyRetriever._class_session = session
        return HTTPSPublicKeyRetriever._class_session

    def retrieve(
        self, key_identifier: Union[KeyIdentifier, str], **requests_kwargs: Any
    ) -> Any:
        """returns the public key for given key_identifier."""
        if not isinstance(key_identifier, KeyIdentifier):
            key_identifier = KeyIdentifier(key_identifier)
        if self._proxies and "proxies" not in requests_kwargs:
            requests_kwargs["proxies"] = self._proxies
        url = self.base_url + key_identifier.key_id
        try:
            return self._retrieve(url, requests_kwargs)
        except requests.RequestException as e:
            try:
                status_code = e.response.status_code if e.response else None
            except AttributeError:
                status_code = None
            raise PublicKeyRetrieverException(e, status_code=status_code)

    def _retrieve(self, url: str, requests_kwargs: Any) -> Any:
        resp = self._session.get(
            url, headers={"accept": PEM_FILE_TYPE}, **requests_kwargs
        )
        resp.raise_for_status()
        self._check_content_type(url, resp.headers["content-type"])
        return resp.text

    def _check_content_type(self, url: str, content_type: str):
        msg = EmailMessage()
        msg["content-type"] = content_type
        media_type = msg.get_content_type()

        if media_type.lower() != PEM_FILE_TYPE.lower():
            raise PublicKeyRetrieverException(
                "Invalid content-type, '%s', for url '%s' ." % (content_type, url)
            )


class HTTPSMultiRepositoryPublicKeyRetriever(BasePublicKeyRetriever):
    """This class retrieves public key from the supplied https key
    repository locations based upon key ids.
    """

    def __init__(self, key_repository_urls: Iterable[str]) -> None:
        if not isinstance(key_repository_urls, list):
            raise TypeError("keystore_urls must be a list of urls.")
        self._retrievers = self._create_retrievers(key_repository_urls)

    def _create_retrievers(
        self, key_repository_urls: Iterable[str]
    ) -> Iterable[BasePublicKeyRetriever]:
        return [HTTPSPublicKeyRetriever(url) for url in key_repository_urls]

    def handle_retrieval_exception(
        self, retriever: BasePublicKeyRetriever, exception: Exception
    ):
        """Handles working with exceptions encountered during key
        retrieval.
        """
        if isinstance(exception, PublicKeyRetrieverException):
            original_exception = getattr(exception, "original_exception", None)
            if isinstance(original_exception, ConnectionError):
                return
            if exception.status_code is None or exception.status_code < 500:
                raise

    def retrieve(
        self, key_identifier: Union[KeyIdentifier, str], **requests_kwargs: Any
    ) -> Any:
        for retriever in self._retrievers:
            try:
                return retriever.retrieve(key_identifier, **requests_kwargs)
            except (RequestException, PublicKeyRetrieverException) as e:
                self.handle_retrieval_exception(retriever, e)
                logger = logging.getLogger(__name__)
                logger.warning(
                    "Unable to retrieve public key from store",
                    extra={
                        "underlying_error": str(e),
                        "key repository": getattr(retriever, "base_url", "unknown"),
                    },
                )
        raise PublicKeyRetrieverException("Cannot load key from key repositories")


class BasePrivateKeyRetriever(object):
    """This is the base private key retriever class."""

    def load(self, issuer: str) -> Tuple[Union[KeyIdentifier], Union[str, bytes]]:
        """returns the key identifier and private key pem found
        for the given issuer.
        """
        raise NotImplementedError("Not implemented.")


class DataUriPrivateKeyRetriever(BasePrivateKeyRetriever):
    """This class can be used to retrieve the key identifier and
    private key from the supplied data uri.
    """

    def __init__(self, data_uri: str) -> None:
        self._data_uri = data_uri

    def load(self, issuer: str) -> Tuple[Union[KeyIdentifier], Union[str, bytes]]:
        if not self._data_uri.startswith("data:application/pkcs8;kid="):
            raise PrivateKeyRetrieverException("Unrecognised data uri format.")
        splitted = self._data_uri.split(";")
        key_identifier = KeyIdentifier(unquote_plus(splitted[1][len("kid=") :]))
        key_data = base64.b64decode(splitted[-1].split(",")[-1])
        key = serialization.load_der_private_key(
            key_data,
            password=None,
            backend=cryptography.hazmat.backends.default_backend(),
        )
        private_key_pem = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return key_identifier, private_key_pem.decode("utf-8")


class StaticPrivateKeyRetriever(BasePrivateKeyRetriever):
    """This class simply returns the key_identifier and private_key_pem
    initially provided to it in calls to load.
    """

    def __init__(
        self,
        key_identifier: Union[KeyIdentifier, str],
        private_key_pem: Union[str, bytes],
    ) -> None:
        if not isinstance(key_identifier, KeyIdentifier):
            key_identifier = KeyIdentifier(key_identifier)

        self.key_identifier: KeyIdentifier = key_identifier
        self.private_key_pem: Union[str, bytes] = private_key_pem

    def load(self, issuer: str) -> Tuple[Union[KeyIdentifier], Union[str, bytes]]:
        return self.key_identifier, self.private_key_pem


class FilePrivateKeyRetriever(BasePrivateKeyRetriever):
    """This class can be used to retrieve the latest key identifier and
    private key for a given issuer found under its private key
    repository path.
    """

    def __init__(self, private_key_repository_path: str) -> None:
        self.private_key_repository = FilePrivateKeyRepository(
            private_key_repository_path
        )

    def load(self, issuer: str) -> Tuple[KeyIdentifier, str]:
        key_identifier = self._find_last_key_id(issuer)
        private_key_pem = self.private_key_repository.load_key(key_identifier)
        return key_identifier, private_key_pem

    def _find_last_key_id(self, issuer) -> KeyIdentifier:
        key_identifiers = list(self.private_key_repository.find_valid_key_ids(issuer))

        if key_identifiers:
            return key_identifiers[-1]
        else:
            raise IOError("Issuer has no valid keys: %s" % issuer)


class FilePrivateKeyRepository(object):
    """This class represents a file backed private key repository."""

    def __init__(self, path) -> None:
        self.path = path

    def find_valid_key_ids(self, issuer: str) -> Generator[KeyIdentifier, Any, None]:
        issuer_directory = os.path.join(self.path, issuer)
        for filename in sorted(os.listdir(issuer_directory)):
            if filename.endswith(".pem"):
                yield KeyIdentifier("%s/%s" % (issuer, filename))

    def load_key(self, key_identifier: KeyIdentifier) -> str:
        key_filename = os.path.join(self.path, key_identifier.key_id)
        with open(key_filename, "rb") as f:
            return f.read().decode("utf-8")



from typing import Any, Iterable, Optional, Protocol, Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, rsa

import atlassian_jwt_auth
from atlassian_jwt_auth import KeyIdentifier
from atlassian_jwt_auth.signer import JWTAuthSigner


def get_new_rsa_private_key_in_pem_format() -> bytes:
    """returns a new rsa key in pem format."""
    private_key = rsa.generate_private_key(
        key_size=2048, backend=default_backend(), public_exponent=65537
    )
    return private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )


def get_public_key_pem_for_private_key_pem(private_key_pem: bytes) -> bytes:
    private_key = serialization.load_pem_private_key(
        private_key_pem, password=None, backend=default_backend()
    )
    public_key = private_key.public_key()
    return public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )


def get_example_jwt_auth_signer(**kwargs: Any) -> JWTAuthSigner:
    """returns an example jwt_auth_signer instance."""
    issuer = kwargs.get("issuer", "egissuer")
    key_id = kwargs.get("key_id", "%s/a" % issuer)
    key = kwargs.get("private_key_pem", get_new_rsa_private_key_in_pem_format())
    algorithm = kwargs.get("algorithm", "RS256")
    return atlassian_jwt_auth.create_signer(issuer, key_id, key, algorithm=algorithm)


def create_token(
    issuer: str,
    audience: Union[str, Iterable[str]],
    key_id: Union[KeyIdentifier, str],
    private_key: Union[str, bytes],
    subject: Optional[str] = None,
):
    """ " returns a token based upon the supplied parameters."""
    signer = atlassian_jwt_auth.create_signer(
        issuer, key_id, private_key, subject=subject
    )
    return signer.generate_jwt(audience)


class BaseJWTAlgorithmTestMixin(object):
    """A mixin class to make testing different support for different
    jwt algorithms easier.
    """

    def get_new_private_key_in_pem_format(self) -> bytes:
        """returns a new private key in pem format."""
        raise NotImplementedError("not implemented.")


class KeyMixInProtocol(Protocol):
    @property
    def algorithm(self) -> str: ...

    def get_new_private_key_in_pem_format(self) -> bytes: ...


class RS256KeyTestMixin(KeyMixInProtocol):
    """Private rs256 test mixin."""

    @property
    def algorithm(self) -> str:
        return "RS256"

    def get_new_private_key_in_pem_format(self):
        return get_new_rsa_private_key_in_pem_format()


class ES256KeyTestMixin(KeyMixInProtocol):
    """Private es256 test mixin."""

    @property
    def algorithm(self) -> str:
        return "ES256"

    def get_new_private_key_in_pem_format(self):
        private_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
        return private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )



from typing import Callable, Iterable, Optional

from atlassian_jwt_auth.frameworks.flask.decorators import with_asap


def requires_asap(
    f: Callable,
    issuers: Optional[Iterable[str]] = None,
    subject_should_match_issuer: Optional[bool] = None,
) -> Callable:
    """
    Wrapper for Flask endpoints to make them require asap authentication to
    access.
    """

    return with_asap(
        func=f,
        required=True,
        issuers=issuers,
        subject_should_match_issuer=subject_should_match_issuer,
    )



from typing import List


def get_permitted_algorithm_names() -> List[str]:
    """returns permitted algorithm names."""
    return [
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512",
    ]



from typing import Any, Optional

from flask import Request, Response, current_app, g
from flask import request as current_req

from ..common.backend import Backend
from ..common.utils import SettingsDict


class FlaskBackend(Backend):
    def get_authorization_header(self, request: Optional[Request] = None) -> bytes:
        if request is None:
            request = current_req

        auth_header = request.headers.get("AUTHORIZATION", "")
        return (
            auth_header.encode("utf-8") if isinstance(auth_header, str) else auth_header
        )

    def get_401_response(
        self,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        request: Optional[Request] = None,
    ) -> Response:
        if headers is None:
            headers = {}

        headers.update(self.default_headers_401)

        return Response(data, status=401, headers=headers)

    def get_403_response(
        self,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        request: Optional[Request] = None,
    ) -> Response:
        return Response(data, status=403, headers=headers)

    def set_asap_claims_for_request(self, request: Request, claims: Any) -> None:
        g.asap_claims = claims

    @property
    def settings(self) -> SettingsDict:
        settings = {}

        settings.update(self.default_settings)

        for k in settings.keys():
            value = current_app.config.get(k)
            if value is None:
                continue

            settings[k] = value

        return self._process_settings(settings)



from collections import OrderedDict
from functools import lru_cache
from typing import Any, Dict, Iterable, Optional, Sequence, Union

import jwt
import jwt.api_jwt
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from jwt import PyJWK
from jwt.exceptions import InvalidAlgorithmError

from atlassian_jwt_auth import KeyIdentifier, algorithms, exceptions, key
from atlassian_jwt_auth.key import BasePublicKeyRetriever

AllowedPublicKeys = Union[
    RSAPublicKey, EllipticCurvePublicKey, Ed25519PublicKey, Ed448PublicKey
]


@lru_cache(maxsize=10)
def _load_public_key(
    algorithms: Sequence[str], public_key: str, algorithm: Optional[str]
) -> Any:
    """Returns a public key object instance given the public key and
    algorithm.

    This has been extracted out of JWTAuthVerifier to avoid possible memory
    leaks via retained instance references.
    """
    if isinstance(public_key, (RSAPublicKey, EllipticCurvePublicKey)):
        return public_key
    if algorithm is None or algorithm not in algorithms:
        raise InvalidAlgorithmError("The specified alg value is not allowed")
    py_jws = jwt.api_jws.PyJWS(algorithms=algorithms)
    alg_obj = py_jws._algorithms[algorithm]
    return alg_obj.prepare_key(public_key)


class JWTAuthVerifier(object):
    """This class can be used to verify a JWT."""

    def __init__(
        self, public_key_retriever: BasePublicKeyRetriever, **kwargs: Any
    ) -> None:
        self.public_key_retriever = public_key_retriever
        self.algorithms = algorithms.get_permitted_algorithm_names()
        self._seen_jti: OrderedDict[str, None] = OrderedDict()
        self._subject_should_match_issuer = kwargs.get(
            "subject_should_match_issuer", True
        )
        self._check_jti_uniqueness = kwargs.get("check_jti_uniqueness", False)

    def verify_jwt(
        self,
        a_jwt: Union[str, bytes],
        audience: str,
        leeway: int = 0,
        **requests_kwargs: Any,
    ) -> Dict[Any, Any]:
        """Verify if the token is correct

        Returns:
             dict: the claims of the given jwt if verification is successful.

        Raises:
            ValueError: if verification failed.
        """
        key_identifier = key._get_key_id_from_jwt_header(a_jwt)
        public_key = self._retrieve_pub_key(key_identifier, requests_kwargs)

        alg = jwt.get_unverified_header(a_jwt).get("alg", None)
        public_key_obj = self._load_public_key(public_key, alg or "RS256")
        return self._decode_jwt(
            a_jwt, key_identifier, public_key_obj, audience=audience, leeway=leeway
        )

    def _retrieve_pub_key(
        self, key_identifier: Union[KeyIdentifier, str], requests_kwargs: Any
    ) -> str:
        return self.public_key_retriever.retrieve(key_identifier, **requests_kwargs)

    def _load_public_key(self, public_key: str, algorithm: Optional[str]) -> Any:
        """Returns a public key object instance given the public key and
        algorithm.
        """
        return _load_public_key(tuple(self.algorithms), public_key, algorithm)

    def _decode_jwt(
        self,
        a_jwt: Union[str, bytes],
        key_identifier: KeyIdentifier,
        jwt_key: Union[AllowedPublicKeys, PyJWK, str, bytes],
        audience: Optional[Union[str, Iterable[str]]] = None,
        leeway: int = 0,
    ) -> Dict[Any, Any]:
        """Decode JWT and check if it's valid"""
        options = {
            "verify_signature": True,
            "require": ["exp", "iat"],
            "require_exp": True,
            "require_iat": True,
        }

        claims = jwt.decode(
            a_jwt,
            key=jwt_key,
            algorithms=self.algorithms,
            options=options,
            audience=audience,
            leeway=leeway,
        )

        if (
            not key_identifier.key_id.startswith("%s/" % claims["iss"])
            and key_identifier.key_id != claims["iss"]
        ):
            raise ValueError("Issuer does not own the supplied public key")

        if self._subject_should_match_issuer and (
            claims.get("sub") and claims["iss"] != claims["sub"]
        ):
            raise exceptions.SubjectDoesNotMatchIssuerException(
                "Issuer does not match the subject."
            )

        _aud = claims.get("aud", None)
        if _aud is None:
            _msg = (
                "Claims validity, the aud claim must be provided and cannot be empty."
            )
            raise KeyError(_msg)
        _exp = int(claims["exp"])
        _iat = int(claims["iat"])
        if _exp - _iat > 3600:
            _msg = "Claims validity, '%s', exceeds the maximum 1 hour." % (_exp - _iat)
            raise ValueError(_msg)
        _jti = claims["jti"]
        if self._check_jti_uniqueness:
            self._check_jti(_jti)
        return claims

    def _check_jti(self, jti: str) -> None:
        """Checks that the given jti has not been already been used."""
        if jti in self._seen_jti:
            raise exceptions.JtiUniquenessException(
                "The jti, '%s', has already been used." % jti
            )
        self._seen_jti[jti] = None
        while len(self._seen_jti) > 1000:
            self._seen_jti.popitem(last=False)
